python test_script/test_jax_cuda_kernel.py
python test_script/test_jax_cuda_kernel_single.py
python test_script/test_jax_cuda_kernel_inference.py

python test_script/test_torch_cuda_kernel_single.py
python test_script/test_torch_cuda_kernel.py
python test_script/test_torch_cuda_kernel_inference.py

python test_script/test_tf_cuda_kernel.py
python test_script/test_tf_cuda_kernel_single.py


#python test_script/test_rwkv6_kernel.py --backend torch --kernel-type cuda
#python test_script/test_rwkv6_kernel.py --backend jax --kernel-type cuda


